local M = {}
local logTag = "hybrid"

-- DEBUG
local _debug = false
local dd = obj.debugDrawProxy
local textCol = color(255, 60, 255, 255)

function dText(row, text)
    if not _debug then
        return
    end

    dd:drawText2D(float3(20, 20 + (20 * row), 0), textCol, text)
end

function dLog(level, text)
    if not _debug then
        return
    end
    
    if not text then -- Shift arguments
        text = level
        level = "D"
    end
    
    log(level, logTag, text)
end

-- END DEBUG

M.type = "auxilliary"
M.relevantDevice = nil
M.defaultOrder = 1000

local enableAutoStop = false

local constants = {
    -- From shiftLogic
    rpmToAV = 0.104719755,
    avToRPM = 9.549296596425384,
    -- Regen braking simulation
    minFriction = 90,
    maxFriction = 5200,
    minRegenSpeed = 4,
    maxRegenSpeed = 22,
    regenThrottleMidpoint = 0.25,
    -- Engine accel delay
    minEngineEngageSpeed = 8,
    maxEngineEngageSpeed = 18,
    minThrusterFadeSpeed = 8,
    maxThrusterFadeSpeed = 35,
    -- Smoothing
    sThrottleIn = 1.65,
    sThrottleOut = 1.85,
    -- Sounds
    inverterBaseVolume = 0.85,
    inverterPitchMinSpeed = 1,
    inverterPitchMaxSpeed = 3
}

local smoothers = {
    throttle = nil
}

local sounds = {
    inverterWhine = nil
}

-- Math helpers
local function clamp(val, min, max)
    return math.min(math.max(val, min), max)
end

local cvt = nil
local engine = nil

local throttle = 0
local effectiveThrottle = 0
local wheelMph = 0
local tractionEffort = 0
local regenEffort = 0
local lastIgnition = false
local lastIgnitionControl = nil
local hasMovedSinceEngineStart = false

local function updateValues(dt)
    throttle = smoothers.throttle:get(input.throttle or 0.0, dt)
    wheelMph = electrics.values.wheelspeed * 2.236
end

local function updateFriction(dt)
    -- Friction override (to simulate dynamic braking)

    if electrics.values.gearIndex > 0 and electrics.values.ignition then
        local speedCoef = (wheelMph - constants.minRegenSpeed) / (constants.maxRegenSpeed - constants.minRegenSpeed)
        local regenCoef = clamp(speedCoef, 0.0, 1.0)
        local actualMidpoint = clamp(constants.regenThrottleMidpoint * regenCoef, 0.01, 1.0)
        local pedalRegenAmount = clamp(actualMidpoint - throttle, 0.0, actualMidpoint) / actualMidpoint
        local actualThrottleAmount = clamp((throttle - actualMidpoint) / (1.0 - actualMidpoint), 0.0, 1.0)
        local newFriction =
            regenCoef * pedalRegenAmount * (constants.maxFriction - constants.minFriction) + constants.minFriction

        dText(0, "Throttle: " .. (throttle * 100) .. "%")

        effectiveThrottle = actualThrottleAmount
        cvt.friction = newFriction
        regenEffort = regenCoef * pedalRegenAmount
    else
        effectiveThrottle = throttle
        cvt.friction = constants.minFriction
        regenEffort = 0.0
    end
end

local function updateThruster(dt)
    -- Engine doesn't kick in right away; some torque comes from electric at first
    -- We use a thruster to simulate the electric part of the drivetrain

    if electrics.values.gearIndex > 0 and electrics.values.ignition then
        local speedCoefEng =
            (wheelMph - constants.minEngineEngageSpeed) /
            (constants.maxEngineEngageSpeed - constants.minEngineEngageSpeed)
        local speedCoefThr =
            (wheelMph - constants.minThrusterFadeSpeed) /
            (constants.maxThrusterFadeSpeed - constants.minThrusterFadeSpeed)
        local engineDelayCoef = clamp(speedCoefEng, 0.05, 1.0)
        local thrusterCoef = throttle * (1.0 - clamp(speedCoefThr, 0.0, 1.0))

        dText(1, "Engine delay %: " .. engineDelayCoef * 100 .. "%")
        dText(2, "Thruster %: " .. thrusterCoef * 100 .. "%")

        electrics.values.throttleOverride = effectiveThrottle * engineDelayCoef
        electrics.values.hybrid_thruster = thrusterCoef
        tractionEffort = thrusterCoef
    elseif electrics.values.gear == "R" and electrics.values.ignition then
        local reverseCoef = throttle * (1.0 - clamp(wheelMph / 12, 0.0, 1.0))

        electrics.values.throttleOverride = 0.0 -- No throttle in reverse
        electrics.values.hybrid_thruster = -reverseCoef
        tractionEffort = reverseCoef
    else
        electrics.values.throttleOverride = effectiveThrottle
        electrics.values.hybrid_thruster = 0.0
        tractionEffort = 0.0
    end
end

local function updateEngine(dt)
    local ignitionControl = electrics.values.ignition
    local autoStop =
        enableAutoStop and 
        hasMovedSinceEngineStart and 
        electrics.values.gearIndex > 0 and 
        input.brake > 0.15 and
        wheelMph < 1
    local ignition = not autoStop

    if ignition and wheelMph > 4 then
        -- Allow engine to shut down again

        hasMovedSinceEngineStart = true
    end
    
    -- Obey actual game ignition control

    if lastIgnitionControl == nil then
        lastIgnitionControl = ignitionControl
    end

    if ignitionControl ~= lastIgnitionControl then
        ignition = ignitionControl
        hasMovedSinceEngineStart = false
    end

    -- Start or stop the engine if needed
    
    if ignition and not lastIgnition then
        engine:activateStarter()
        engine:setIgnition(ignition and 1 or 0)
        hasMovedSinceEngineStart = false
    elseif not ignition then
        engine:setIgnition(0)
        engine:deactivateStarter()
    end

    lastIgnition = ignition
    lastIgnitionControl = ignitionControl
end

local function updateSounds()
    -- Update sounds
    local inverterPitchCoef =
        (wheelMph - constants.inverterPitchMinSpeed) /
        (constants.inverterPitchMaxSpeed - constants.inverterPitchMinSpeed)
    local inverterEffort = math.max(tractionEffort, regenEffort)
    local volume = constants.inverterBaseVolume * clamp(inverterEffort, 0.0, 1.0)
    local pitch = clamp(inverterPitchCoef, 0.4, 1.0)

    obj:setVolumePitch(sounds.inverterWhine, volume, pitch)
end

local function initSounds()
    sounds.inverterWhine = obj:createSFXSource(
        "vehicles/citybus/sounds/inverter.ogg",
        "AudioDefaultLoop3D",
        "inverterWhine",
        engine.engineNodeID
    )
end

local function init()
    smoothers.throttle = newTemporalSmoothingNonLinear(constants.sThrottleIn, constants.sThrottleOut)
end

local function initSecondStage()
    cvt = powertrain.getDevice("gearbox")
    engine = powertrain.getDevice("mainEngine")
end

local function updateGFX(dt)
    updateValues(dt)

    if cvt then
        updateFriction(dt)
        updateThruster(dt)
    end

    if engine then
        updateEngine(dt)
    end

    updateSounds()
end

M.init = init
M.initSounds = initSounds
M.initSecondStage = initSecondStage
M.updateGFX = updateGFX

return M
